(ns kube-deploy.core
  (:require [lambdakube.core :as lk]
            [kube-deploy.helpers :as lkh]
            [lambdakube.util :as lku]
            [clojure.java.shell :as sh]
            [clojure.java.io :as io]))

(def conf (atom {}))
(def config {:num-be-slaves 3 :num-fe-replicas 3})

(def resources-mini
  {:requests
   {:cpu "100m" :memory "64Mi"}
   :limits
   {:cpu "200m" :memory "256Mi"}})

(def resources-small
  {:requests
   {:cpu "100m" :memory "100Mi"}
   :limits
   {:cpu "200m" :memory "400Mi"}})

(def resources-medium
  {:requests
   {:cpu "200m" :memory "400Mi"}
   :limits
   {:cpu "500m" :memory "1Gi"}})

; will not be needed soon
(defn kube-apply-namespace
  ([content file]
   (kube-apply-namespace content file nil nil))
  ([content file namespace]
   (kube-apply-namespace content namespace nil))
  ([content file kube-namespace kube-config]
   (let [namespace-param (if (nil? namespace) nil (str "--namespace=" kube-namespace))
         config-param (if (nil? config) nil (str "--kubeconfig=" kube-config))]
     (when-not (and (.exists file)
                    (= (slurp file) content))
       (spit file content)
       (let [res (apply sh/sh (remove nil? ["kubectl" "apply" namespace-param config-param "-f" (str file)]))]
         (when-not (= (:exit res) 0)
           (.delete file)
           (throw (Exception. (:err res)))))))))

(defn redis [kube-injector]
  (-> kube-injector
      (lk/rule :redis-master-rule []
               (fn []
                 (-> (lk/pod :redis-master {:app :redis
                                            :role :master
                                            :tier :backend})
                     (lk/add-container :redis (get-in @conf [:images :redis])
                                       (-> {}
                                           (lkh/add-to-spec :resources resources-small)
                                           (lkh/add-readiness-probe-cmd ["redis-cli" "--version"])
                                           (lkh/add-liveness-probe-cmd ["redis-cli" "--version"])))
                     (lk/deployment 1)
                     (lk/expose-cluster-ip :redis-master
                                           (lk/port :redis :redis 6379 6379)))))

      (lk/rule :redis-slave-rule [:redis-master-rule :num-be-slaves]
               (fn [backend-master num-be-slaves]
                 (-> (lk/pod :redis-slave {:app :redis
                                           :role :slave
                                           :tier :backend})
                     (lk/add-container :redis (get-in @conf [:images :redis])
                                       (-> {}
                                           (lkh/add-to-spec :resources resources-small)
                                           (lkh/add-to-spec
                                            :args ["--slaveof"
                                                   (:hostname backend-master)
                                                   (-> backend-master :ports :redis str)])
                                           (lkh/add-liveness-probe-cmd ["redis-cli" "--version"])))
                     (lk/deployment num-be-slaves)
                     (lk/expose-cluster-ip :redis-slave
                                           (lk/port :redis :redis 6379 6379)))))))

(defn database [kube-injector]
  (-> kube-injector
      (lk/rule :postgres-master-rule []
               (fn []
                 (-> (lk/pod
                      :postgres-master
                      {:app :postgres :role :master :tier :backend})
                     (lk/add-container :postgres (get-in @conf [:images :postgres])
                                       (-> {}
                                           (lkh/add-to-spec :resources resources-small)
                                           (lkh/add-readiness-probe-cmd ["pg_isready"])
                                           (lkh/add-liveness-probe-cmd ["pg_isready"])
                                           (lkh/add-to-spec :envFrom [{:secretRef {:name "environment"}}])))

                     (lk/deployment 1)
                     (lk/expose-cluster-ip :postgres-master
                                           (lk/port :postgres :postgres 5432 5432)))))))

(defn rabbitmq [kube-injector]
  (-> kube-injector
      (lk/rule :rabbit-master-rule []
               (fn []
                 (-> (lk/pod
                      :rabbit-master
                      {:app :rabbit :role :master :tier :backend})

                     (lk/add-container
                      :rabbit
                      (get-in @conf [:images :rabbitmq])
                      (-> {}
                          (lkh/add-to-spec :resources resources-medium)
                          (lkh/add-liveness-probe-cmd ["rabbitmq-diagnostics" "ping" "-q"])))

                     (lk/deployment 1)
                     (lk/expose-cluster-ip
                      :rabbit-master (lk/port :rabbit :rabbit 5672 5672)))))))

(defn hackspace [kube-injector]
  (-> kube-injector
      (lk/rule :hackspace-backend-master [:postgres-master-rule :redis-master-rule]
               (fn [postgres redis]
                 (-> (lk/pod :hackspace-master {:app :hackspace
                                                :role :master
                                                :tier :backend}
                             {:imagePullSecrets [{:name "docker-registry"}]})

                     ; Add containers to the pod, one for frontend assets and one for backend
                     (lk/add-container
                      :hackspace-backend
                      (get-in @conf [:images :django])
                      (-> {}
                          (lkh/add-to-spec :resources resources-medium)
                          (lkh/add-to-spec :command ["/gunicorn.sh"])
                          (lkh/add-to-spec :ports [{:containerPort 5000 :name "gunicorn-port"}])
                          (lkh/add-to-spec :envFrom [{:secretRef {:name "environment"}}])
                          (lkh/add-readiness-probe-cmd ["wget" "127.0.0.1:5000"])
                          (lkh/add-liveness-probe-cmd ["wget" "127.0.0.1:5000"])))

                     #_(lk/add-container
                      :hackspace-frontend
                      (get-in @conf [:images :nginx])
                      (-> {}
                          (lkh/add-to-spec :resources resources-mini)
                          (lkh/add-to-spec :ports [{:containerPort 80 :name "nginx-port"}])
                          (lkh/add-readiness-probe-cmd ["wget" "127.0.0.1/static/images/favicon.ico"])
                          (lkh/add-liveness-probe-cmd ["wget" "127.0.0.1/static/images/favicon.ico"])))
                     (lk/add-volume "static-files" {:emptyDir {}}
                                    {:hackspace-backend "/app/staticfiles"})

                       ; order does matter here, deployment after initcontainers
                     (lku/wait-for-service-port postgres :postgres)
                     (lku/wait-for-service-port redis :redis)

                       ; wrap in a deployment
                     (lk/deployment 2)
                     (lk/expose-cluster-ip :hackspace-backend
                                           (lk/port :hackspace-master :gunicorn-port 5000 5000))
                     (lk/expose-cluster-ip :hackspace-frontend
                                           (lk/port :hackspace-master :nginx-port 80 80)))))
      (lk/rule :hackspace-queue-master []
               (fn []
                 (-> (lk/pod :hackspace-queue
                             {:app :hackspace
                              :role :master
                              :tier :backend}
                             {:imagePullSecrets [{:name "docker-registry"}]})
                     (lk/add-container
                      :hackspace
                      (get-in @conf [:images :django])
                      (-> {}
                          (lkh/add-to-spec :resources resources-small)
                          (lkh/add-to-spec :command ["/app/manage.py" "huey"])
                          (lkh/add-to-spec :envFrom [{:secretRef {:name "environment"}}])))
                     (lk/deployment 2))))))

(defn footprint [kube-injector]
  (-> kube-injector
      (lk/rule :footprint-backend-master [:postgres-master-rule :redis-master-rule :rabbit-master-rule]
               (fn [postgres redis rabbit]
                 (-> (lk/pod :footprint-master {:app :footprint
                                                :role :master
                                                :tier :backend}
                             {:imagePullSecrets [{:name "docker-registry"}]})

                     ; Add containers to the pod, one for frontend assets and one for backend
                     (lk/add-container
                      :footprint-backend
                      (get-in @conf [:images :footprint])
                      (-> {}
                          (lkh/add-to-spec :resources resources-medium)
                          (lkh/add-to-spec :command ["/gunicorn.sh"])
                          (lkh/add-to-spec :ports [{:containerPort 5000 :name "gunicorn-port"}])
                          (lkh/add-to-spec :envFrom [{:secretRef {:name "environment"}}])
                          #_(lkh/add-readiness-probe-get "http://127.0.0.1" 5000)
                          (lkh/add-liveness-probe-get "http://127.0.0.1" 5000)))

                     (lk/add-container
                      :footprint-frontend
                      (get-in @conf [:images :nginx])
                      (-> {}
                          (lkh/add-to-spec :resources resources-mini)
                          (lkh/add-to-spec :ports [{:containerPort 80 :name "nginx-port"}])
                          #_(lkh/add-readiness-probe-get "127.0.0.1/static/images/favicon.ico" 80)
                          (lkh/add-liveness-probe-get "http://127.0.0.1/static/images/favicon.ico" 80)))

                     (lk/add-volume "static-files" {:emptyDir {}}
                                    {:footprint-backend "/app/staticfiles"
                                     :footprint-frontend "/usr/share/nginx/html/static/"})

                     ; order does matter here, deployment after initcontainers
                     (lku/wait-for-service-port postgres :postgres)
                     (lku/wait-for-service-port redis :redis)
                     (lku/wait-for-service-port rabbit :rabbit)

                     ; wrap in a deployment
                     (lk/deployment 2)
                     (lk/expose-cluster-ip :footprint-backend
                                           (lk/port :footprint-master :gunicorn-port 5000 5000))
                     (lk/expose-cluster-ip :footprint-frontend
                                           (lk/port :footprint-master :nginx-port 80 80)))))
      (lk/rule :footprint-queue-master []
               (fn []
                 (-> (lk/pod :footprint-queue
                             {:app :footprint
                              :role :master
                              :tier :backend}
                             {:imagePullSecrets [{:name "docker-registry"}]})

                     (lk/add-container
                      :footprint
                      (get-in @conf [:images :footprint])
                      (-> {}
                          (lkh/add-to-spec :resources resources-small)
                          (lkh/add-to-spec :command ["/app/manage.py" "rundramatiq" "-p" "1" "-t" "1"])
                          (lkh/add-to-spec :envFrom [{:secretRef {:name "environment"}}])))

                     (lk/deployment 2))))))

(defn make-secrets [kube-injector name data]
  (-> kube-injector
      (lk/rule :secrets []
               (fn []
                 (-> (lkh/secrets (keyword name) data))))))

(defn make-registry-secrets [kube-injector name data]
  (-> kube-injector
      (lk/rule
       :secrets-registry []
       (fn []
         (lkh/secrets-registry name {} (lkh/encode-secret-map data))))))

(defn make-namespace [kube-injector name]
  (-> kube-injector
      (lk/rule :namespace []
               (fn []
                 (lkh/kube-namespace (keyword name))))))

(defn make-environment-secrets [kube-injector name data]
  (-> kube-injector
      (lk/rule :secrets-environment []
               (fn []
                 (lkh/secrets name {}
                              (lkh/encode-secret-map data))))))

(defn make-ingress [kube-injector]
  (lk/rule kube-injector :lkube-ingress []
           (fn []
             (-> (lkh/ingress "lkube-ingress" {:app :nginx-ingress})
                 (lkh/add-spec :tls [{:hosts [:lkube-ingress] :secretName "lkube-tls"}])
                 (lkh/add-spec :rules [{:host "lkube-ingress.35.197.252.20.nip.io"
                                        :http {:paths
                                               [{:path "/"
                                                 :backend
                                                 {:serviceName "footprint-backend"
                                                  :servicePort 5000}}
                                                {:path "/static/"
                                                 :backend
                                                 {:serviceName "footprint-frontend"
                                                  :servicePort 80}}]}}])
                 (lk/add-annotation :kubernetes.io/tls-acme "true")
                 (lk/add-annotation :kubernetes.io/ingress.class "nginx")))))

(defn footprint-deployment [namespace kube-config]
  (reset! conf (clojure.edn/read-string (slurp (str namespace "-config.edn"))))
  ; run deployment in default namespace to create a namespace called footprint
  (-> (lk/injector)
      (make-namespace namespace)
      lk/standard-descs
      (lk/get-deployable config)
      (lkh/to-yaml-store (str namespace "-store.yaml"))
      (kube-apply-namespace (io/file (str namespace "-deploy.yaml")) nil kube-config))

  ; run deployment under footprint namespace
  (-> (lk/injector)
      (make-environment-secrets :environment (:vars @conf))
      (make-registry-secrets :docker-registry {:.dockerconfigjson (:registry @conf)})
      make-ingress
      rabbitmq
      database
      redis
      footprint
      lk/standard-descs
      (lk/get-deployable config)
      (lkh/to-yaml-store (str namespace "-store.yaml"))
      (kube-apply-namespace (io/file (str namespace "-deploy.yaml")) "footprint" kube-config)))

(defn hackspace-deployment [namespace kube-config]
  (reset! conf (clojure.edn/read-string (slurp (str namespace "-config.edn"))))
  ; run deployment in default namespace to create a namespace called footprint
  (-> (lk/injector)
      (make-namespace namespace)
      ;lk/standard-descs
      (lk/get-deployable config)
      (lkh/to-yaml-store (str namespace "-store.yaml"))
      (kube-apply-namespace (io/file (str namespace "-deploy.yaml")) nil kube-config))

  ; run deployment under footprint namespace
  (-> (lk/injector)
      (make-environment-secrets :environment (:vars @conf))
      ;(make-registry-secrets :docker-registry {:.dockerconfigjson (:registry @conf)})
      make-ingress
      database
      redis
      hackspace
      lk/standard-descs
      (lk/get-deployable config)
      (lkh/to-yaml-store (str namespace "-store.yaml"))
      (kube-apply-namespace (io/file (str namespace "-deploy.yaml")) namespace kube-config)))

(defn -main []
  (hackspace-deployment "maidstone-hackspace" "/home/oly/.kube/dke.yml")
  (footprint-deployment "footprint" "/home/oly/.kube/ake.yml"))

